from typing import List

import logging
import yfinance as yf
from tqdm import tqdm
import pandas as pd

from .base import ColumnNameHandler


def load_daily_returns(list_of_tickers: List[str], min_date: str, max_date: str) -> pd.DataFrame:
    """
    Function to load daily return for the specified list of tickers and dates from Yahoo Finance.

    Parameters
    ----------
    list_of_tickers : List[str]
        List of tickers to load
    min_date : str
        Date to load the returns from
    max_date : str
        Date to load the return untill

    Returns
    -------
    pd.DataFrame
        Dataframe with ticker, Ret and Date columns
    """
    returns_df = list()
    # Keep only one-token tickers (otherwise, will fall with error) and keep only unique
    list_of_tickers = set([t for t in list_of_tickers if len(t.split())==1])
    for t in tqdm(list_of_tickers):
        # Load stock data daily
        ticker_df = yf.download(t, min_date, max_date, progress=False, auto_adjust=False, multi_level_index=False)
        # If no data available - skip the ticker
        if 'Adj Close' not in ticker_df or ticker_df.empty:
            print(f"Could not load return for ticker {t}. Skipping it.")
            continue
        
        # Calculate daily returns as a percentage change
        ticker_df.loc[:, ColumnNameHandler.ret_col] = ticker_df['Adj Close'].pct_change()
        # Remove Nan values
        ticker_df.dropna(inplace=True)
        ticker_df.reset_index(inplace=True)
        # Append ticker name as a new column
        ticker_df.loc[:, ColumnNameHandler.ticker_col] = t
        returns_df.append(ticker_df[[
            ColumnNameHandler.ticker_col, 
            ColumnNameHandler.date_col, 
            ColumnNameHandler.ret_col, 
            ColumnNameHandler.volume_col
            ]].copy())
    if not returns_df:
        return None
    # Combine all returns together
    returns_df = pd.concat(returns_df, axis=0)
    returns_df.dropna(axis=1, inplace=True)
    returns_df.reset_index(inplace=True, drop=True)
    returns_df[ColumnNameHandler.date_col] = pd.to_datetime(returns_df[ColumnNameHandler.date_col]).dt.date
    return returns_df


def load_daily_factors(five_factors: bool = False) -> pd.DataFrame:
    """
    Function to load daily factors from Fama-Franch public library.

    Parameters
    ----------
    five_factors : bool, optional
        Whether to load five factors or three factors instead, by default False

    Returns
    -------
    pd.DataFrame
        Dataframe with loaded factors per date
    """
    website_url = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"
    if five_factors:
        url_to_factors = f'{website_url}/F-F_Research_Data_5_Factors_2x3_daily_CSV.zip'
    else:
        url_to_factors = f'{website_url}/F-F_Research_Data_Factors_daily_CSV.zip'
    # First 3 rows show meta data - thus, skip it
    ff_df = pd.read_csv(url_to_factors, skiprows=3)
    # Original index is Date - thus, in the file it will be unnamed. Change it.
    ff_df.rename({'Unnamed: 0': 'Date'}, axis=1, inplace=True)
    ff_df.loc[:, 'Date'] = pd.to_datetime(ff_df['Date'].astype(str), errors='coerce').dt.date
    # There could be a footer with metadata - make sure to drop it
    ff_df.dropna(axis=0, inplace=True)
    return ff_df



from typing import Any, Iterable, Callable
import pandas as pd
import numpy as np
import datetime as dt
from scipy.stats import norm
import plotly.graph_objects as go
from joblib import Parallel, delayed

from .base import ColumnNameHandler


def expand_date_columns(input_df: pd.DataFrame, date_col: str, before=400, after=30) -> pd.DataFrame:
    """
    Expand each date with n days before and n days after

    Parameters
    ----------
    input_df : pd.DataFrame
        Input dataframe
    date_col : str
        Date column
    before : int, optional
        Number of days to add before the date, by default 400
    after : _type_, optional
        Number of days to add after the date, by default 30

    Returns
    -------
    pd.DataFrame
        Dataframe with exploded dates
    """
    input_df.loc[:, date_col] = pd.to_datetime(input_df[date_col]).dt.date
    expand_date_mapping = {
        d: [d + dt.timedelta(days=int(i)) for i in np.arange(-before, after+1)] 
        for d in input_df[date_col].unique()
    }
    input_df.loc[:, date_col] = input_df[date_col].map(expand_date_mapping)
    input_df = input_df.explode(date_col)
    input_df.reset_index(drop=True, inplace=True)
    return input_df


def calculate_car_stats(event_res_df: pd.DataFrame, critical_value: float = 0.95) -> pd.DataFrame:
    """
    Calculate to statistics for mean CAR effect with confidence levels over the event window
    Parameters
    ----------
    event_res_df : pd.DataFrame
        Result dataframe from the output of event-study. Should contain event_id, offset, and car columns.
    critical_value : float, optional
        Critical value to use in confidence intervals estimation, by default 0.95

    Returns
    -------
    pd.DataFrame
        Dataframe with estimated statistics
    """
    car_df = event_res_df.pivot(
        columns=ColumnNameHandler.offset_col, 
        index=ColumnNameHandler.event_id_col, 
        values=ColumnNameHandler.car_col)
    # Calculate mean and STD per each offset-day
    stat_car_df = pd.concat([
        car_df.mean().to_frame('mean'),
        car_df.median().to_frame('median'),
        # STD for each offset day as the deviation from mean CAR on the day, divided N * (N-1), where N - number of asset on the day
        np.sqrt(((car_df - car_df.mean())**2).sum() / (car_df.notna().sum() * (car_df.notna().sum() - 1))).to_frame('sd')
        ], axis=1)
    # Add T-stat results
    stat_car_df.loc[:, 't_stat'] = stat_car_df['mean'] / (stat_car_df['sd'])
    stat_car_df.loc[:, 'p_value'] = stat_car_df['t_stat'].apply(lambda v: norm.cdf(-abs(v)) * 2).round(3)
    # Add confidence levels
    conf_level = - norm.ppf((1-critical_value) / 2)
    stat_car_df.loc[:, 'upper_ci'] = stat_car_df['mean'] + conf_level * stat_car_df['sd']
    stat_car_df.loc[:, 'lower_ci'] = stat_car_df['mean'] - conf_level * stat_car_df['sd']
    return stat_car_df


def plot_mean_car(event_res_df: pd.DataFrame, critical_value: float = 0.95, 
                  color_rgb: str = '0, 0, 255', bootstrap_df: pd.DataFrame = None) -> go.Figure:
    """
    Function to plot the mean CAR effect with confidence levels over the event window
    Parameters
    ----------
    event_res_df : pd.DataFrame
        Result dataframe from the output of event-study. Should contain event_id, offset, and car columns.
    critical_value : float, optional
        Critical value to use in confidence intervals estimation, by default 0.95
    color_rgb: str
        String specifying rgb color to be used in the plot. Default, blue (0,0,255)
    bootstrap_df: pd.DataFrame, optional
        Dataframe with bootstrap results to be used in confidence level identification

    Returns
    -------
    go.Figure
        Plotly figure
    """
    plot_df = calculate_car_stats(event_res_df=event_res_df, critical_value=critical_value)
    
    single_tail_critical_value = (1-critical_value)/2
    conf_level = np.round(- norm.ppf(single_tail_critical_value), 2)
    upper_conf_name = f'Mean + {conf_level}SE'
    lower_conf_name = f'Mean - {conf_level}SE'

    if bootstrap_df is not None:
        single_tail_critical_value = (1-critical_value)/2
        confidence_df = bootstrap_df.T.quantile([single_tail_critical_value, 1-single_tail_critical_value]).T
        confidence_df.columns = ['lower_ci', 'upper_ci']
        plot_df.drop(['lower_ci', 'upper_ci'], axis=1, inplace=True)
        plot_df = plot_df.join(confidence_df)
        upper_conf_name = f'Q:{round(1-single_tail_critical_value, 3)} by bootstrap'
        lower_conf_name = f'Q:{round(single_tail_critical_value, 3)} by bootstrap'

    # Add Mean Trace
    trace_one = go.Scatter(
        x=plot_df.index,
        y=plot_df['mean'],
        name='Mean',
        showlegend=True,
        mode='lines',
        line=dict(color=f'rgb({color_rgb})')
    )
    # Add trace for the upper bound
    trace_fill_one = go.Scatter(
        x=plot_df.index,
        y=plot_df['upper_ci'],
        name=upper_conf_name,
        mode='lines',
        line=dict(width=1, color=f'rgba({color_rgb}, 0.5)', dash='dash'),
        showlegend=True,
        
    )
    # Add trace for the lower bound
    trace_fill_two = go.Scatter(
        x=plot_df.index,
        y=plot_df['lower_ci'],
        name=lower_conf_name,
        mode='lines',
        line=dict(width=1, color=f'rgba({color_rgb}, 0.5)', dash='dash'),
        showlegend=True,
    )
    # Add trace for the lower bound
    trace_fill_two_shade = go.Scatter(
        x=plot_df.index,
        y=plot_df['lower_ci'],
        fill='tonextx',
        name='ShadedArea',
        mode='lines',
        fillcolor=f'rgba({color_rgb},0.1)',
        line=dict(width=1, color=f'rgba({color_rgb}, 0.5)', dash='dash'),
        showlegend=False,
    )
    # Update layout  
    layout = go.Layout(
        title=f"""<b>Cumulitative Abnormal Return: Mean & {round(critical_value*100)}% Confidence Limits</b><br>"""\
        f"""<i>Based on {event_res_df[ColumnNameHandler.event_id_col].nunique()} non-missing events""",
        template='plotly_white',
        xaxis_title='Day Relative to Event',
        yaxis_title='Return',
        yaxis_tickformat=',.1%',
        legend=dict(
                orientation="h",
                xanchor = "center",
                x=0.5,
                y=-0.2
        )        
    )
    fig = go.Figure(data=[
        trace_one, 
        trace_fill_one,
        trace_fill_two_shade,
        trace_fill_two
        ], layout=layout
    )
    fig.add_vline(
        x=0, 
        line_width=1, 
        line_color='rgba(0, 0,0, 0.5)')
    fig.add_hline(
        y=0, 
        line_width=1, 
        line_color='rgba(0, 0,0, 0.5)')
    return fig


def plot_joint_mean_car(event_res_one: pd.DataFrame, event_res_two: pd.DataFrame, 
                        name_one: str, name_two: str,  critical_value: float = 0.95) -> go.Figure:
    """
    Function to plot two mean CAR effects with confidence levels ont the same plot.
    
    Parameters
    ----------
    event_res_one : pd.DataFrame
        First result dataframe from the output of event-study. Should contain event_id, offset, and car columns.
    event_res_two : pd.DataFrame
        Second result dataframe from the output of event-study. Should contain event_id, offset, and car columns.
    name_one : str
        Name of the first event study
    name_two : str
        Name of the second event study
    critical_value : float, optional
        Critical value to use in confidence intervals estimation, by default 0.95
    Returns
    -------
    go.Figure
        Plotly figure
    """
    fig_one = plot_mean_car(event_res_one)
    fig_two = plot_mean_car(event_res_two, color_rgb='0, 153, 0')
    for trace in fig_one.data:
        trace.legendgroup = 'group1'
        trace.legendgrouptitle=dict(text=f'<b>{name_one}</b> ({event_res_one[ColumnNameHandler.event_id_col].nunique()} events)')

    for trace in fig_two.data:
        trace.legendgroup = 'group2'
        trace.legendgrouptitle=dict(text=f'<b>{name_two}</b> ({event_res_two[ColumnNameHandler.event_id_col].nunique()} events)')

    for trace in fig_one.data:
        fig_two.add_trace(trace)

    fig_two.update_layout(
        title=f"""<b>Cumulitative Abnormal Return: Mean & {round(critical_value*100)}% Confidence Limits</b><br>""",
        legend=dict(
            tracegroupgap=50,
            orientation="h",
            xanchor = "center",
            x=0.55,
            y=-0.3,
            entrywidth=200,
            traceorder='grouped'
        )
    )
    return fig_two


def run_in_parallel(func: Callable, data: Iterable[Any], n_cores: int) -> Iterable[Any]:
    if n_cores == 1:
        return [func(x) for x in data]
    processed_data = Parallel(n_jobs=n_cores)(
        delayed(func)(x) for x in data
    )
    return processed_data



from typing import List

import pandas as pd
from pandas.core.api import DataFrame as DataFrame
import numpy as np
import statsmodels.api as sm
from sklearn.base import BaseEstimator, RegressorMixin

from .base import LogMixin, ColumnNameHandler


class BaseEstimator(BaseEstimator, RegressorMixin, LogMixin, ColumnNameHandler):
    def __init__(self, feature_cols: List[str]):
        self.feature_cols = feature_cols

    def fit(self, x: pd.DataFrame, y=None) -> 'BaseEstimator':
        return self
    
    def predict(self, x: pd.DataFrame) -> List[float]:
        raise NotImplementedError
    
    @property
    def resid(self) -> np.ndarray:
        raise NotImplementedError
    

class RegReturnEstimator(BaseEstimator):
    def __init__(self, feature_cols: List[str]):
        self.estimator = None
        self.feature_cols = feature_cols if isinstance(feature_cols, List) else [feature_cols]
    
    def fit(self, x: DataFrame, y=None) -> BaseEstimator:
        self.estimator = sm.OLS(
            x[self.ret_col], 
            sm.add_constant(x[self.feature_cols])
        ).fit()
        return self
    
    def predict(self, x: DataFrame) -> List[float]:
        return self.estimator.predict(sm.add_constant(x[self.feature_cols])).tolist()
    
    @property
    def resid(self) -> np.ndarray:
        if self.estimator:
            return self.estimator.resid
        return
    

class MAMReturnEstimator(BaseEstimator):
    def __init__(self):
        self.__resid = None
    
    @property
    def resid(self) -> float:
        return self.__resid

    def fit(self, x: pd.DataFrame, y=None) -> 'MAMReturnEstimator':
        self.__resid = (x[self.ret_col] - x[self.mkt_rf_col] - x[self.rf_col])**2
        return self
    
    def predict(self, x: pd.DataFrame) -> List[float]:
        return (x[self.mkt_rf_col] + x[self.rf_col]).tolist()



import logging
import pandas as pd
from dataclasses import dataclass
from sklearn.base import BaseEstimator, RegressorMixin


class LogMixin(object):
    @property
    def logger(self):
        name = '.'.join([__name__, self.__class__.__name__])
        return logging.getLogger(name)


@dataclass
class ColumnNameHandler:
    ret_col: str = 'Ret'
    offset_col: str = 'Offset'
    ref_rf_col: str = 'Ret-RF'
    rf_col: str = 'RF'
    mkt_rf_col: str = 'Mkt-RF'
    smb_col: str = 'SMB'
    hml_col: str = 'HML'
    rmw_col: str = 'RMW'
    cma_col: str = 'CMA'
    date_col: str = 'Date'
    ticker_col: str = 'ticker'
    volume_col: str = 'Volume'
    event_date_col: str = 'event_date'
    event_id_col: str = 'event_id'

    # Predicted cols
    pred_ret_col: str = 'pred_ret_col'
    ar_col: str = 'ar'
    car_col: str = 'car'
    sar_col: str = 'sar'
    scar_col: str = 'scar'
    e_std_col: str = 'estimation_std'


class BasePandasRegressor(BaseEstimator, RegressorMixin, LogMixin, ColumnNameHandler):
    def fit(self, x: pd.DataFrame, y=None):
        return self
    def predict(self, x: pd.DataFrame, y=None) -> pd.DataFrame:
        raise NotImplementedError
    def fit_predict(self, x: pd.DataFrame, y=None) -> pd.DataFrame:
        return self.fit(x, y).predict(x, y)
    


from typing import List, Union, Dict

import pandas as pd
from copy import deepcopy
from dateutil import parser
from tqdm import tqdm
import numpy as np
from datetime import timedelta, datetime

from .base import ColumnNameHandler as CNH
from .data_loader import load_daily_returns, load_daily_factors
from .utils import expand_date_columns, run_in_parallel, calculate_car_stats
from .estimation_models import RegReturnEstimator, MAMReturnEstimator, BaseEstimator


def apply_estimator(input_dict: Dict[str, Union[BaseEstimator, pd.DataFrame]]) -> pd.DataFrame:
    """
    Function to apply estimator for a given input.
    Separated from the class for a multiprocessing purposes

    Parameters
    ----------
    input_dict : Dict[str, Union[BaseEstimator, pd.DataFrame]]
        Dictionary with input arguments. Should contains keys: estimator, estimation_df, prediction_df

    Returns
    -------
    pd.DataFrame
        Dataframe with predicted returns
    """
    estimator = input_dict['estimator']
    estimation_df = input_dict['estimation_df']
    prediction_df = input_dict['prediction_df']

    estimator.fit(estimation_df)
    resid_std = estimator.resid.std()
    prediction_df.loc[:, CNH.pred_ret_col] = estimator.predict(prediction_df)
    # Get all abnormal columns
    prediction_df.sort_values(CNH.offset_col, inplace=True)
    prediction_df.loc[:, CNH.ar_col] = prediction_df[CNH.ret_col] - prediction_df[CNH.pred_ret_col]
    prediction_df.loc[:, CNH.car_col] = prediction_df[CNH.ar_col].cumsum()
    prediction_df.loc[:, CNH.sar_col] = prediction_df[CNH.ar_col] / resid_std
    prediction_df.loc[:, CNH.scar_col] = prediction_df[CNH.car_col] / (
        resid_std * np.sqrt(prediction_df.reset_index(drop=True).index+1))
    prediction_df.loc[:, CNH.e_std_col] = resid_std
    return prediction_df


class EventStudy(CNH):
    AUX_OFFSET: int = 10
    def __init__(self, estimation_days: int = 255, gap_days: int = 50, 
                 window_before: int = 10, window_after: int = 10, 
                 min_estimation_days: int = 100, estimator_type: str = 'capm',
                 n_cores: int = None, verbose: bool = False):
        self.estimation_days = estimation_days
        self.gap_days = gap_days
        self.window_before = window_before
        self.window_after = window_after
        self.min_estimation_days = min_estimation_days
        self.returns_df = None
        self.factors_df = None
        self.estimator_type = estimator_type
        self.n_cores = n_cores
        self.estimator = self.__initialize_estimator()
        self.verbose = verbose

    def print(self, msg: str):
        if self.verbose:
            print(f"{datetime.now().time().strftime('%H:%M:%S')} - {msg}")
    
    def __initialize_estimator(self):
        if self.estimator_type.lower() == 'capm':
            return RegReturnEstimator(feature_cols=self.mkt_rf_col)
        elif self.estimator_type.lower() == 'ff3':
            return RegReturnEstimator(feature_cols=[self.mkt_rf_col, self.smb_col, self.hml_col])
        elif self.estimator_type.lower() == 'ff5':
            return RegReturnEstimator(feature_cols=[self.mkt_rf_col, self.smb_col, self.hml_col, self.rmw_col, self.cma_col])
        elif self.estimator_type.lower() == 'mam':
            return MAMReturnEstimator()
        else:
            raise NotImplementedError(f"Estimator type {self.estimator_type} is not implemented.")
    
    def add_factors(self, factors_df: pd.DataFrame=None):
        if factors_df is not None:
            self.factors_df = factors_df
            return
        if self.estimator_type == 'ff5':
            self.factors_df = load_daily_factors(five_factors=True)
        else:
            self.factors_df = load_daily_factors()
        # Need to downscale factor columns from percent to float
        factor_names = self.factors_df.columns.difference([self.date_col])
        self.factors_df.loc[:, factor_names] = self.factors_df[factor_names]/100
        return
        
    def process_events_df(self, input_df: pd.DataFrame) -> pd.DataFrame:
        """
        Function to process input events. In particular:
        1. Remove duplicates
        2. Convert dates into an appropriate datetime format
        3. Add auxiliary columns like offset for easier estimation/event window access
        Parameters
        ----------
        input_df : pd.DataFrame
            Input datatframe with ticker and event_date columns

        Returns
        -------
        pd.DataFrame
            Process event dataframe 
        """
        event_df = input_df[[self.ticker_col, self.event_date_col]].copy()
        event_df[self.event_date_col] = pd.to_datetime(event_df[self.event_date_col]).dt.date
        # Add event ID column
        event_df.loc[:, self.event_id_col] = event_df.reset_index(drop=True).index
        # Remove duplicates and convert to datetime
        event_df.drop_duplicates(inplace=True)
        event_df.loc[:, self.date_col] = event_df[self.event_date_col].copy()
        # Adjust day of the event if needed
        self.print(f"Expanding date column...")
        event_df = expand_date_columns(
            event_df, 
            date_col=self.date_col,
            before=self.window_before+self.gap_days + self.estimation_days + self.AUX_OFFSET,
            after=self.window_after + self.AUX_OFFSET)
        event_df[self.date_col] = pd.to_datetime(event_df[self.date_col]).dt.date
        return event_df
    
    def add_returns(self, list_of_tickers: List[str]=None, 
                    min_date: str=None, max_date: str=None, 
                    ret_df: pd.DataFrame=None):
        """
        Function to load returns for specified tickers and time period.
        Parameters
        ----------
        list_of_tickers : List[str], optional
            List of tickers to load returns for, by default None
        min_date : str, optional
            Minimum date to load the returns from, by default None
        max_date : str, optional
            Maximum date to load the returns until, by default None
        ret_df : pd.DataFrame, optional
            Data frame of pre-loaded returns, by default None
        """
        # If pre-loaded returns are provided - use them
        if ret_df is not None:
            ret_df.loc[:, self.date_col] = pd.to_datetime(ret_df[self.date_col]).dt.date
            self.returns_df = ret_df[[self.ret_col, self.date_col, self.ticker_col, self.volume_col]]
            return 
        if list_of_tickers is None or min_date is None or max_date is None:
            raise ValueError('Either returns DF should be provided, or the loading parameters.')
        
        # Add offset to the dates to make sure all returns for all specified dates are returned
        if isinstance(min_date, str):
            min_date = parser.parse(min_date)
        if isinstance(max_date, str):
            max_date = parser.parse(max_date)
        min_date = min_date - timedelta(days=1)
        max_date = max_date + timedelta(days=1)
        self.returns_df = load_daily_returns(
            list_of_tickers=list_of_tickers,
            min_date=min_date, 
            max_date=max_date
        )
        return
    
    def __add_offset(self, input_df: pd.DataFrame) -> pd.DataFrame:
        """
        Function to create an offset column. Offset - how many days is between an event and a particular date.
        Note - it should be done for trading and not calendar days. Thus, only after the inner-join with returns data

        Parameters
        ----------
        input_df : pd.DataFrame
            Event dataframe to process

        Returns
        -------
        _type_
            _description_
        """
        # Lets create a column with name offSet but fill at first with order number by event_id
        input_df.sort_values([self.event_id_col, self.date_col], inplace=True)
        input_df.loc[:, self.offset_col] = input_df.groupby(self.event_id_col).cumcount()
        # Get mask associated with event itself
        event_mask = input_df[self.event_date_col]==input_df[self.date_col]
        # For each event_id - get its offset value
        event_to_offset_dict = input_df[event_mask].set_index(self.event_id_col)[self.offset_col].to_dict()
        # Remove event day value to obtain valid offset
        input_df.loc[:, self.offset_col] = input_df[self.offset_col] - input_df[self.event_id_col].map(event_to_offset_dict)
        input_df.dropna(inplace=True)
        return input_df

    def run_study(self, x: pd.DataFrame) -> pd.DataFrame:
        self.print(f"Processing input events...")
        event_df = self.process_events_df(x)

        # Adding Returns data
        if self.returns_df is None:
            self.print(f"Start loading the returns...")
            self.add_returns(
                event_df[self.ticker_col].unique(), 
                min_date=event_df[self.date_col].min(),
                max_date=event_df[self.date_col].max()
            )
        self.print(f"Merging events and returns data...")
        event_df = pd.merge(
            event_df,
            self.returns_df,
            on=[self.ticker_col, self.date_col], how='outer'
        )
        event_df.dropna(inplace=True)

        # Add offset column
        self.print(f"Creating event offset column...")
        event_df = self.__add_offset(event_df)
        
        # Add factors data
        self.print(f"Adding factors data...")
        if self.factors_df is None:
            self.add_factors()
        event_df = pd.merge(event_df, self.factors_df, on=self.date_col)
        
        # Create estimation and prediction periods
        estimation_period = np.arange(
            -(self.window_before+self.gap_days+self.estimation_days), 
            -(self.window_before+self.gap_days), 
            1)
        prediction_period = np.arange(-self.window_before, self.window_after + 1, 1)

        self.print(f"Initializing single event-studies...")
        # Create chunks by event_id
        groups = [
            {'estimator': deepcopy(self.estimator),
             'estimation_df':  group[group[self.offset_col].isin(estimation_period)],
             'prediction_df':  group[group[self.offset_col].isin(prediction_period)]}
             for _, group in event_df.groupby(self.event_id_col)
        ]
        # Filter out events with lower number of estimation observations
        groups = [g for g in groups if g['estimation_df'].shape[0] >= self.min_estimation_days]
        self.print(f"Running event-study for all single events...")
        prediction_df = run_in_parallel(
            apply_estimator,
            groups, 
            n_cores=self.n_cores
        ) 
        prediction_df = pd.concat(prediction_df, ignore_index=True)
        return prediction_df

    def run_bootstrap_study(self, x: pd.DataFrame, n_bootstrap: int = 10) -> pd.DataFrame:
        """
        Run bootstrap event studies based on input events. 
        The function shuffles input events n_bootstrap times, and calculates mean CAR on each run.

        Parameters
        ----------
        x : pd.DataFrame
            Input events dataframe with ticker and event_date column
        n_bootstrap : int, optional
            Number of bootstrap runs, by default 10

        Returns
        -------
        pd.DataFrame
            Dataframe with mean CAR values for each bootstrap run
        """
        n_events = x.drop_duplicates().shape[0]
        self.print(f"Creating bootstrap input events...")
        options = list()
        for _ in tqdm(range(n_bootstrap)):
            opt_df = x.copy()
            opt_df.loc[:, self.event_date_col] = x[self.event_date_col].sample(frac=1).tolist()
            options.append(opt_df.copy())
        bootstrap_df = pd.concat(options, axis=0)
        res_df = self.run_study(bootstrap_df)
        self.print(f"Calculating mean CAR for bootstrap runs...")
        mean_car_list = list()
        for i in range(0, n_bootstrap):
            mask = res_df[self.event_id_col].between(i*n_events, (i+1) * n_events,inclusive='left')
            mean_car_list.append(
                calculate_car_stats(res_df[mask])[['mean']].rename({'mean': i}, axis=1).copy())
        return pd.concat(mean_car_list, axis=1)
